#include "oop.h"
#include "minmax.h"

#include <string.h>
#include <assert.h>

String::String(int size) : ByteArray(size + 1)
{
  mElements[mSize - 1]= '\0';
}

String::String(char *string) : ByteArray(strlen(string) + 1)
{
  memcpy(mElements, string, mSize);
  assert(0 == mElements[mSize - 1]);
}

oop String::classNew(int size)
{
  return new String(size);
}

unsigned String::hash(void)
{
  unsigned h= 0;
  for (int index= 0;  index < mSize;  ++index)
    if (mElements[index])
      h= (h & 0x1fffffff) * 31 + mElements[index];
  return h;
}

int String::compare(oop anObject)
{
  if (anObject->isString())
    return strcmp((char *)mElements, anObject->cString());
  int len1= size();
  int len2= anObject->size();
  for (int i= 0;  i < ::min(len1, len2);  ++i)
    {
      oop c1= at(i);
      oop c2= anObject->at(i);
      if (!(*c1 == c2))
	return (*c1 < c2) ? -1 : 1;
    }
  return (len1 == len2) ? 0 : ((len1 < len2) ? -1 : 1);
}

int String::operator <(oop other)
{
  if (!other->isString()) return 0;
  return compare(other) < 0;
}

int String::operator <=(oop other)
{
  if (!other->isString()) return 0;
  return compare(other) <= 0;
}

int String::operator ==(oop other)
{
  if (!other->isString()) return 0;
  return compare(other) == 0;
}

int String::operator >=(oop other)
{
  if (!other->isString()) return 0;
  return compare(other) >= 0;
}

int String::operator >(oop other)
{
  if (!other->isString()) return 0;
  return compare(other) > 0;
}

oop String::operator +(oop other)
{
  int len1= size();
  int len2= other->size();
  char *buf= (char *)alloca(len1 + len2 + 1);
  memcpy(buf, cString(), len1);
  memcpy(buf + len1, other->cString(), len2);
  buf[len1 + len2]= '\0';
  return new String(buf);
}

oop String::at(int index)
{
  if (index < 0 || index >= mSize - 1)
    errorSubscriptBounds(index);
  return Character::value(ByteArray::byteAt(index));
}

oop String::atPut(int index, oop aCharacter)
{
  if (index < 0 || index >= mSize - 1)
    errorSubscriptBounds(index);
  ByteArray::byteAtPut(index, aCharacter->asciiValue());
  return aCharacter;
}

int String::size(void)
{
  return mSize - 1;
}

char *String::cString(void)
{
  return (char *)mElements;
}

oop String::printOn(oop aStream)
{
  aStream->nextPut(Character::value('"'));
  aStream->nextPutAll(this);
  aStream->nextPut(Character::value('"'));
  return this;
}
